/*

VSEN block cipher

Uses 32x4 dynamic PRNG-based key-dependent substitution boxes
and PRNG-based self-preencrypted extended rounds keys,
core round function is like a one from GOST,
but with unique sboxes for each round and unique round keys.

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define KEY_LEN 64
// Key length in bytes, 16-64

#define PRNG_CYCLES 1024
#define PRNG_CYCLES_ST 2048
// Mix iterations

#define FIRST 1
#define SECOND 0

typedef unsigned char u8;
typedef unsigned long int u32;
typedef signed long int s32;

#include "rnd.h"
#include "init.h"
#include "key.h"

void __stdcall crypt (u32 *);
void __stdcall decrypt (u32 *);


static u8 GENBOX[128][256];
static u8 RBOX[32][4][256];
static u8 KBOX[256];
static u32 EXKEY[32];



static u8
generate_box (const u8 ss, const u8 * key, u32 keylen, u32 first)
{
  u32 id, i, cycles;
  u8 t, s;
  u8 *GB;
  const u8 *SB;
  const u8 *IB;

  id = 0;

  s = ss;

  for (id = 0; id != 128; id++)
    {

      GB = &GENBOX[id][0];
      SB = &SBOX[id][0];
      IB = &IBOX[id][0];


      if (first)
	{
	  cycles = PRNG_CYCLES;

	  for (i = 0; i != 256; i++)
	    {
	      GB[i] = IB[i];
	    }

	}
      else
	{
	  cycles = PRNG_CYCLES_ST;

	  for (i = 0; i != 256; i++)
	    {
	      GB[i] = IB[GB[i] ^ i];
	    }

	}

      for (i = 0; i != cycles; i++)
	{
	  s = GB[(s + SB[(GB[i % 256] + key[i % keylen]) % 256]) % 256];
	  t = GB[i % 256];
	  GB[i % 256] = GB[s % 256];
	  GB[s % 256] = t;
	}
    }


  return s;
}



static u8
generate_key_box (const u8 ss, const u8 * key, u32 keylen, u32 first)
{
  u32 i, cycles;
  u8 t, s;
  u8 *GB;
  const u8 *SB;
  const u8 *IB;

  s = ss;

  GB = &KBOX[0];
  SB = &SKBOX[0];
  IB = &IKBOX[0];

  if (first)
    {
      cycles = PRNG_CYCLES;

      for (i = 0; i != 256; i++)
	{
	  GB[i] = IB[i];
	}

    }
  else
    {
      cycles = PRNG_CYCLES_ST;

      for (i = 0; i != 256; i++)
	{
	  GB[i] = (u8) i;
	}

    }

  for (i = 0; i != cycles; i++)
    {
      s = GB[(s + SB[(GB[i % 256] + key[i % keylen]) % 256]) % 256];
      t = GB[i % 256];
      GB[i % 256] = GB[s % 256];
      GB[s % 256] = t;
    }


  return s;
}




static u32
f (const u32 x, const u32 r)
{

  return (_lrotl
	  (RBOX[r][0][(x) >> 24 & 255] << 24 | RBOX[r][1][(x) >> 16 & 255] <<
	   16 | RBOX[r][2][(x) >> 8 & 255] << 8 | RBOX[r][3][(x) & 255], 11));

}





void __stdcall __export
setup (const u8 * userkey)
{
  u32 i, n, id, sub;
  u8 *RSB, *mat;
  u32 pos;
  u8 s;
  u8 key[128];
  u8 *ek;

  for (i = 0; i != KEY_LEN; i++)
    {
      key[i] = userkey[i];
    }


  s = generate_box (0, &key[0], KEY_LEN, FIRST);

  mat = &GENBOX[0][0];

  pos = 0;


  for (id = 0; id != 32; id++)
    {
      for (sub = 0; sub != 4; sub++)
	{
	  RSB = &RBOX[id][sub][0];

	  for (n = 0; n != 256; n++)
	    {
	      RSB[n] = mat[pos++];
	    }

	}


    }


  s = generate_key_box (s, key, KEY_LEN, FIRST);

  ek = (u8 *) & EXKEY[0];

  for (i = 0; i != 128; i++)
    {
      ek[i] = KBOX[128 + i];
    }

  ek = (u8 *) & KBOX[0];

  for (i = 0; i != 128; i += 8)
    {
      ek += i;
      crypt ((u32 *) ek);
    }


  ek = (u8 *) & EXKEY[0];

  for (i = 0; i != 128; i++)
    {
      ek[i] = KBOX[i];
      key[i] = KBOX[i];
    }


  generate_box (s, &key[0], 128, SECOND);

  mat = &GENBOX[0][0];

  pos = 0;


  for (id = 0; id != 32; id++)
    {
      for (sub = 0; sub != 4; sub++)
	{
	  RSB = &RBOX[id][sub][0];

	  for (n = 0; n != 256; n++)
	    {
	      RSB[n] = mat[pos++];
	    }

	}
    }

  s = generate_key_box (s, key, 128, SECOND);

  ek = (u8 *) & EXKEY[0];

  for (i = 0; i != 128; i++)
    {
      ek[i] = KBOX[128 + i];
    }

  ek = (u8 *) & KBOX[0];

  for (i = 0; i != 128; i += 8)
    {
      ek += i;
      decrypt ((u32 *) ek);
    }


  ek = (u8 *) & EXKEY[0];

  for (i = 0; i != 128; i++)
    {
      ek[i] = KBOX[i];
    }

}



void __stdcall __export
crypt (u32 * in)
{
  register u32 n1, n2;
  u32 r;

  n1 = in[0];
  n2 = in[1];

  r = 0;

  do
    {
      n2 ^= f (n1 + EXKEY[r], r);
      r++;
      n1 ^= f (n2 + EXKEY[r], r);
      r++;
    }

  while (r != 32);

  in[0] = n2;
  in[1] = n1;
}



void __stdcall __export
decrypt (u32 * in)
{
  register u32 n1, n2;
  s32 r;

  n2 = in[0];
  n1 = in[1];

  r = 31;

  do
    {
      n1 ^= f (n2 + EXKEY[r], r);
      r--;
      n2 ^= f (n1 + EXKEY[r], r);
      r--;
    }

  while (r != -1);

  in[0] = n1;
  in[1] = n2;
}



u32 __stdcall __export
getblocksize ()
{
  return 64;
}

u32 __stdcall __export
getkeysize ()
{
  return KEY_LEN * 8;
}

void __stdcall __export
getciphername (u8 * p)
{
  strcpy (p, "VSEN");
}
